import os
import sys
import random
import errno
import json
import time

from efs.common.encryption import encrypt, decrypt, hash, import_key, generate_key, pubkey_to_symmetric_key

class Node(object):
    
    def __init__(self, structure, id, **kwargs):
        self.structure = structure
        self.conn = structure.connection
        self.parent = None
        self.name = None
        self.id = id
        self.owner = self.conn.user
        self.group = self.conn.user

        for key in kwargs:
            self.__dict__[key] = kwargs[key]

    def to_dict(self):
        return {"type": self.__class__.__name__,
                "name": self.name,
                "id": self.id,
                "owner": self.owner,
                "group": self.group,
                "parent_id": self.parent.id if self.parent is not None else "-1"}

    @classmethod
    def from_dict(cls, d, structure):
        if "type" not in d:
            return None
        if d["type"] == "File":
            return File.from_dict(d, structure)
        elif d["type"] == "Directory":
            return Directory.from_dict(d, structure)

    def delete(self):
        """Delete this file or directory"""
        self.parent._delete_child(self.name)
        self.conn.rmnode(self.id)

    def move(self, new_path):
        """Rename or move this file or directory"""
        self.parent._move_child(self.name, new_path)
        self.conn.mv(self.id, self.parent.id)

    def get_path(self):
        return self.parent.get_path() + self.name

    def is_file(self):
        return False

    def is_dir(self):
        return False


class File(Node):

    def __init__(self, structure, id, last_mod=0, **kwargs):
        super(File, self).__init__(structure, id, **kwargs)

        self.last_mod = last_mod
        
        self._modified = False
        self._buffer = None
        self._key = None
        self._keys = None

    def to_dict(self):
        d = super(File, self).to_dict()
        d["last_mod"] = self.last_mod
        return d

    @classmethod
    def from_dict(cls, d, structure):
        return cls(structure, **d)

    def read(self, size, offset=0):
        """Read size bytes, starting from offset"""
        return self._get_contents()[offset : offset + size]

    def write(self, data, offset=0):
        """Write data, starting from offset into the file"""
        self._modified = True
        self._set_contents(self._get_contents()[:offset] + data)

    def __len__(self):
        return len(self._get_contents())

    def _get_contents(self):
        if self._buffer is None:
            encrypted_plus_keys = json.loads(self.conn.read(self.id))

            keys = encrypted_plus_keys["keys"]
            if self.owner == self.conn.user:
                key = decrypt(keys["owner"],
                              pubkey_to_symmetric_key(self.conn.private_key))
            else:
                key = decrypt(keys["group"],
                              pubkey_to_symmetric_key(self.conn.get_decrypted_group_private_key(self.group)))
            self._key = key
            self._keys = keys
            
            unencrypted = json.loads(decrypt(encrypted_plus_keys["encrypted"], key))

            # Don't let the server modify the data (or the other checks)
            if hash(unencrypted["data"]) != unencrypted["hash"]:
                raise SecurityError("Server returned modified file")
            
            data = json.loads(unencrypted["data"])

            # Don't let the server return an old version of the file
            if data["last_mod"] < self.last_mod:
                raise SecurityError("Server returned old file version")

            # Don't let the server return a different file
            if data["id"] != self.id:
                raise SecurityError("Server returned incorrect file")
            
            self._buffer = data["contents"].encode('iso-8859-1')
        return self._buffer

    def _set_contents(self, contents):
        self._buffer = contents
    
    def save(self):
        """Save this file to the server"""
        if self._modified:
            self.last_mod = time.time()
            data = json.dumps({"contents": self._get_contents().decode('iso-8859-1'),
                               "last_mod": self.last_mod,
                               "id": self.id})
            hashed = hash(data)

            if self._key is None:
                # We're saving the file for the first time, so we're
                # guaranteed to be the owner of the file and have the
                # private keys of both the owner and the group
                
                self._key = generate_key()

                self._keys = {"owner": encrypt(self._key, pubkey_to_symmetric_key(self.conn.private_key)),
                              "group": encrypt(self._key, pubkey_to_symmetric_key(self.conn.get_decrypted_group_private_key(self.group)))}

            encrypted = encrypt(json.dumps({"data": data,
                                            "hash": hashed}), self._key)

            encrypted_plus_keys = json.dumps({"encrypted": encrypted,
                                              "keys": self._keys})
            
            self.conn.write(self.id, encrypted_plus_keys)
            self._modified = False
            self.structure.save()

    def chgrp(self, new_group):
        self.group = new_group

        # Add the file to the group structure file
        self.structure.groups[new_group].tree_nodes[self.id] = self

        # Generate a new key and encrypt it with the new group's key
        self._key = None
        self._modified = True
        self.save()
        
        self.conn.chgroup(self.id, new_group)

    def __repr__(self):
        return "<File %s>" % self.get_path()

    def is_file(self):
        return True


class Directory(Node):
    
    def __init__(self, structure, id, **kwargs):
        super(Directory, self).__init__(structure, id, **kwargs)

        self.children = {}

    def to_dict(self):
        d = super(Directory, self).to_dict()
        d["children"] = dict([(name, child.to_dict()) for name, child in self.children.items()])
        return d

    @classmethod
    def from_dict(cls, d, structure):
        dir = cls(structure, **d)
        dir.children = {}
        for name, child in d["children"].items():
            dir.children[name] = Node.from_dict(child, structure)
            dir.children[name].parent = dir
        return dir

    def to_flat_dict(self):
        d = super(Directory, self).to_dict()
        d["flat_children"] = dict([(name, child.id) for name, child in self.children.items()])
        return d

    def delete(self):
        """Delete this directory if it is empty"""
        if len(self.children) == 0:
            super(Directory, self).delete()
        else:
            raise FilesysError(errno.ENOTEMPTY)

    def get(self, name):
        if name == "":
            return self
        if "/" not in name and name in self.children:
            return self.children[name]
        
        name, sep, rest = name.partition("/")

        if name == "":
            return self._get_root().get(rest)
        if name == ".":
            return self.get(rest)
        if name == "..":
            return self.parent.get(rest)
        if name in self.children and self.children[name].is_dir():
            return self.children[name].get(rest)
        
        return None

    def create_file(self, path):
        """Create a file with the given name"""
        file = self._create_node(path, File)
        file._buffer = ""
        file._modified = True
        file.save()
        return file

    def create_dir(self, path):
        """Create a subdirectory with the given name"""
        dir = self._create_node(path, Directory)
        return dir

    def _create_node(self, path, cls):
        dest_dir, name = self._split_path(path)
        
        if name in dest_dir.children:
            raise FilesysError(errno.EEXIST)

        id = self.conn.mknode(dest_dir.id, "600") # TODO: change the perms
        
        dest_dir.children[name] = cls(self.structure, id, parent=dest_dir, name=name)
        self.structure.save()

        return dest_dir.children[name]

    def get_path(self):
        path = ""
        if self.parent is not None and self.parent != self:
            path = self.parent.get_path()
        return path + self.name + "/"

    def __repr__(self):
        return "<Directory %s>" % self.get_path()

    def _get_root(self):
        if self.parent is not None and self.parent != self:
            return self.parent._get_root()
        return self

    def _split_path(self, path):
        """
        Split path into a directory and a name.  If the directory does
        not exist, raise an exception
        """
        dir, sep, name = path.rpartition("/")
        dir = self.get(dir + sep)

        if dir is None:
            raise FilesysError(errno.ENOENT)
        if not dir.is_dir():
            raise FilesysError(errno.ENOTDIR)

        return dir, name

    def _delete_child(self, name):
        assert name in self.children
        del self.children[name]
        self.structure.save()

    def _move_child(self, name, new_path):
        assert name in self.children
        child = self.children[name]

        dest_dir, new_name = self._split_path(new_path)

        if new_name in [".", ".."]:
            raise FilesysError(errno.ENOTEMPTY)

        if new_name in dest_dir.children:
            existing = dest_dir.children[new_name]
            if child == existing:
                # noop
                return
            if existing.is_dir() and len(existing.children) > 0:
                # Don't overwrite a nonempty existing directory
                raise FilesysError(errno.ENOTEMPTY)
        else:
            existing = None
            
        dest_dir.children[new_name] = child
        del self.children[name]
        dest_dir.children[new_name].name = new_name
        dest_dir.children[new_name].parent = dest_dir
        if existing is not None:
            self.conn.rmnode(existing.id)

        self.structure.save()

    def get_descendents(self):
        for child in self.children.values():
            yield child
            if child.is_dir():
                for d in child.get_descendents():
                    yield d

    def is_dir(self):
        return True


class FilesysError(Exception):

    def __init__(self, err):
        Exception.__init__(self, os.strerror(err))
        self.err = err

class SecurityError(Exception):
    pass
