#!/usr/bin/python

import logging
from errno import *
from stat import S_IFDIR, S_IFREG
import sys
import os

from fuse import FUSE, FuseOSError, Operations, LoggingMixIn

from efs.client import files
from efs.client.structure_file import StructureFile
from efs.client.connection import Connection

class EfsFuseOperations(LoggingMixIn, Operations):

    def __init__(self, root):
        self.root = root
        self.fds = {}

    def _next_fd(self):
        for n in xrange(1, 10000):
            if n not in self.fds:
                return n
        raise FuseOSError(ENFILE)

    def _get_file_or_dir(self, path, fh=None):
        if fh is None or fh == 0:
            file = self.root.get(path)
            if file is None:
                raise FuseOSError(ENOENT)
            return file
        else:
            if fh not in self.fds:
                raise FuseOSError(EBADF)
            assert self.fds[fh] is not None
            return self.fds[fh]

    def _get_file(self, path, fh=None):
        file = self._get_file_or_dir(path, fh)
        if not file.is_file():
            raise FuseOSError(EISDIR)
        return file
    
    def _get_dir(self, path, fh=None):
        file = self._get_file_or_dir(path, fh)
        if not file.is_dir():
            raise FuseOSError(ENOTDIR)
        return file

    def create(self, path, mode):
        try:
            self.root.create_file(path)

            fd = self._next_fd()
            self.fds[fd] = self.root.get(path)
            assert self.fds[fd] is not None
            return fd
        except files.FilesysError as e:
            raise FuseOSError(e.err)

    def mkdir(self, path, mode):
        try:
            self.root.create_dir(path)
        except files.FilesysError as e:
            raise FuseOSError(e.err)

    def getattr(self, path, fh=None):
        try:
            file = self._get_file_or_dir(path, fh)

            res = {"st_uid": os.getuid(),
                   "st_gid": os.getgid()}

            if file.is_dir():
                res["st_mode"] = S_IFDIR
                res["st_nlink"] = 2 + len([d for d in file.children.values() if d.is_dir()])

            if file.is_file():
                res["st_mode"] = S_IFREG
                res["st_nlink"] = 1 if file.parent is not None else 0
                res["st_size"] = len(file)

            return res
        except files.FilesysError as e:
            raise FuseOSError(e.err)

    def open(self, path, flags):
        try:
            # TODO: check and remember flags
            file = self.root.get(path)
            if file is None:
                self.root.create_file(path)
                file = self.root.get(path)

            fd = self._next_fd()
            self.fds[fd] = file
            return fd
        except files.FilesysError as e:
            raise FuseOSError(e.err)

    def read(self, path, size, offset, fh):
        try:
            file = self._get_file(path, fh)
            
            return file.read(size, offset)
        except files.FilesysError as e:
            raise FuseOSError(e.err)

    def write(self, path, data, offset, fh):
        try:
            file = self._get_file(path, fh)
            
            file.write(data, offset)
            return len(data)
        except files.FilesysError as e:
            raise FuseOSError(e.err)

    def truncate(self, path, length, fh=None):
        self.write(path, "", length, fh)

    def fsync(self, path, datasync, fh):
        try:
            file = self._get_file(path, fh)

            file.save()
        except files.FilesysError as e:
            raise FuseOSError(e.err)
    
    def release(self, path, fh):
        self.fsync(path, 0, fh)
        del self.fds[fh]

    def readdir(self, path, fh):
        try:
            file = self._get_dir(path, fh)

            return ['.', '..'] + file.children.keys()
        except files.FilesysError as e:
            raise FuseOSError(e.err)

    def rename(self, old, new):
        try:
            file = self.root.get(old)
            # TODO: check that the new path isn't stupid
            file.move(new)
        except files.FilesysError as e:
            raise FuseOSError(e.err)

    def rmdir(self, path):
        try:
            file = self._get_dir(path)
            file.delete()
        except files.FilesysError as e:
            raise FuseOSError(e.err)

    def unlink(self, path):
        try:
            file = self._get_file(path)
            file.delete()
        except files.FilesysError as e:
            raise FuseOSError(e.err)

def mount(root, mountpoint):
    try:
        os.mkdir(mountpoint)
    except OSError:
        pass

    FUSE(EfsFuseOperations(root), mountpoint, foreground=True)

if __name__ == '__main__':
    if len(sys.argv) != 2:
        print('usage: %s MOUNTPOINT ' % sys.argv[0])
        exit(1)

    logging.getLogger().setLevel(logging.DEBUG)
    logging.getLogger().addHandler(logging.StreamHandler(sys.stderr))
    connection = Connection("", "")
    structure = StructureFile(connection, fetch_data=False)
    root = files.Directory(structure, None, "", 0)
    root.parent = root
    structure.root = root
    structure.version = 0

    mount(root, sys.argv[1])
