import json

from efs.client.files import Node, Directory, SecurityError
from efs.common.encryption import hash, encrypt, decrypt, pubkey_to_symmetric_key

class StructureFile(object):
    
    def __init__(self, connection, fetch_data=True):
        self.connection = connection
        self.user = connection.user

        if fetch_data:
            wrapper = json.loads(decrypt(connection.get_user_struct(),
                                         pubkey_to_symmetric_key(
                                             self.connection.private_key)))
            data = unwrap(wrapper)
            
            self.root = Node.from_dict(data["files"], self)
            self.version = data["version"]

            # Merge the group structure files together
            
            # Group structure files aren't part of the user structure
            # file, but we do this here for convenience

            existing_node_ids = set([node.id for node in self.root.get_descendents()])
            
            self.groups = {}
            nodes = {}
            for group in self.connection.get_groups():
                self.groups[group] = GroupStructureFile(self, group)
                for id in self.groups[group].nodes:
                    nodes[id] = self.groups[group].nodes[id]

            # Fine the top-level nodes
            for id in nodes:
                if id in existing_node_ids:
                    # This node is in our user structure file
                    del nodes[id]
                    continue
                node = nodes[id]
                if "flat_children" in node:
                    node["children"] = {}
                    for name, child_id in node["flat_children"].items():
                        if child_id in nodes:
                            node["children"][name] = nodes[child_id]
                            del nodes[child_id]
                        else:
                            # We don't have permission to see this child
                            node["children"][name] = {}

            # Construct a Node for every top-level node
            for id in nodes:
                if nodes[id]["parent_id"] == root.id and nodes[id]["owner"] == nodes[id]["name"]:
                    node = Node.from_dict(nodes[id], self)
                    node.parent = self.root
                    self.root.children[node.name] = node

            # Store references to these Nodes in the GroupStructureFiles
            for node in self.root.get_descendents():
                if node.group in self.groups:
                    group = self.groups[node.group]
                    if node.id in group.tree_nodes:
                        group.tree_nodes[node.id] = node

        else:
            self.root = Directory(self, "0", name="")
            self.root.parent = self.root
            self.version = 0
            self.groups = {}

    def to_encrypted_json(self):
        data = json.dumps({"files": self.root.to_dict(),
                           "version": self.version})
        return encrypt(json.dumps({"data": data,
                                   "hash": hash(data)}),
                       pubkey_to_symmetric_key(self.connection.private_key))

    def save(self):
        self.connection.update_user_struct(self.to_encrypted_json())
        for group in self.groups:
            group.save()

class GroupStructureFile(object):

    def __init__(self, user_structure, group, fetch_data=True):
        self.parent = user_struct
        self.connection = self.parent.connection
        self.group = group

        if fetch_data:
            wrapper = json.loads(decrypt(self.connection.get_group_struct(group),
                                         pubkey_to_symmetric_key(
                                             self.connection.get_decrypted_group_private_key(group))))
            data = unwrap(wrapper)

            # dict from node ids to flat dicts
            self.nodes = data["nodes"]
            self.version = data["version"]

            self.tree_nodes = {}
            for id in self.nodes:
                self.tree_nodes[id] = None
        else:
            self.nodes = {}
            self.version = {}
            self.tree_nodes = {}

    def to_encrypted_json(self):
        for id in self.tree_nodes:
            if self.tree_nodes[id] is not None:
                if self.tree_nodes[id].group != self.group and id in self.nodes:
                    del self.nodes[id]
                else:
                    self.nodes[id] = self.tree_nodes[id].to_flat_dict()
        data = json.dumps({"nodes": self.nodes,
                           "version": self.version})
        return encrypt(json.dumps({"data": data,
                                   "hash": hash(data)}),
                       pubkey_to_symmetric_key(
                           self.connection.get_decrypted_group_private_key(self.group)))

    def save(self):
        self.connection.update_group_struct(self.group, self.to_encrypted_json())

    

def unwrap(wrapper):
    if hash(wrapper["data"]) != wrapper["hash"]:
        raise SecurityError("Server modified structure file")
    return json.loads(wrapper["data"])
