#! /usr/bin/env python

import os, binascii, json
from efs.common.exceptions import InvalidArgsError
from efs.common.exceptions import CorruptedDatabaseError
from efs.server.exceptions import PermissionDeniedError

R = "4"
W = "2"
RW = "6"

mongo_client = None
login_db = None
files_table = None
users_table = None
groups_table = None

def gen_id():
	return binascii.b2a_hex(os.urandom(16))

def check_perms(node, user, req_perms):
	global users_table

	if not node:
		return False

	rec = users_table.find({"user": user})

	if rec.count() != 1:
		return False

	# TODO: make this work
	#user_groups = rec[0]["groups"]
	user_groups = []

	if node["owner"] == user:
		node_perms = node["perms"][0]
	elif node["group"] in user_groups:
		node_perms = node["perms"][1]
	else:
		node_perms = node["perms"][2]

	if req_perms == R:
		if node_perms == R or node_perms == RW:
			return True
		else:
			return False
	elif req_perms == W or req_perms == RW:
		if node_perms == RW:
			return True
		else:
			return False
	else:
		return False

	return False

def lookup_node(node_id):
	global files_table

	rec = files_table.find({"node_id": node_id})

	if rec.count() != 1:
		return None

	return rec[0]

def create_node(user, group, perms, par_id):
	global files_table

	node_id = gen_id()

	files_table.insert({"node_id": node_id, "perms": perms, "owner": user, "group": group, "entries": [], "par_id": par_id})

	parent = lookup_node(par_id)

	if parent == None:
		raise InvalidArgsError

	entries = parent["entries"] + [node_id]

	files_table.update({"node_id": par_id}, {"$set": {"entries": entries}})

	return node_id

def mknode(user, par_id, perms):
	parent = lookup_node(par_id)

	if parent == None:
		raise InvalidArgsError

	if check_perms(parent, user, RW):
		node_id = create_node(user, user, perms, par_id)
		return node_id
	else:
		raise PermissionDeniedError

def rmnode(user, node_id):
	node = lookup_node(node_id)

	if node == None:
		raise InvalidArgsError

	par_id = node["par_id"]
	parent = lookup_node(node["par_id"])

	if node == None:
		raise CorruptedDatabaseError

	if check_perms(parent, user, W):
		files_table.remove({"node_id": node_id})

		entries = parent["entries"]
		entries.remove(node_id)
		files_table.update({"node_id": par_id}, {"$set": {"entries": entries}})
	else:
		raise PermissionDeniedError

def printtree(node_id, indent=0):
	node = lookup_node(node_id)

	if node == None:
		raise CorruptedDatabaseError

	print "	" * indent + node["node_id"]
	for e in node["entries"]:
		printtree(e, indent+1)

def read_node(user, node_id):
	node = lookup_node(node_id)

	if node == None:
		raise InvalidArgsError

	if check_perms(node, user, R):
		f = open(node_id, "r")
		contents = f.read()

		return contents
	else:
		raise PermissionDeniedError

def write_node(user, node_id, contents):
	node = lookup_node(node_id)

	if node == None:
		raise InvalidArgsError

	if check_perms(node, user, W):
		f = open(node_id, "w")
		f.write(contents)
		f.close()
	else:
		raise PermissionDeniedError

def chmod(user, node_id, perms):
	node = lookup_node(node_id)

	if node == None:
		raise InvalidArgsError

	if node["owner"] == user:
		files_table.update({"node_id": node_id}, {"$set": {"perms": perms}})
	else:
		raise PermissionDeniedError

def chgrp(user, node_id, group):
	node = lookup_node(node_id)

	if node == None:
		raise InvalidArgsError

	if node["owner"] == user:
		files_table.update({"node_id": node_id}, {"$set": {"group": group}})
	else:
		raise PermissionDeniedError

def mv_node(user, node_id, new_par_id):
	node = lookup_node(node_id)

	old_par_id = node["par_id"]
	old_parent = lookup_node(old_par_id)

	if old_parent == None:
		raise CorruptedDatabaseError

	new_parent = lookup_node(new_par_id)

	if node == None or new_parent == None:
		raise InvalidArgsError

	if check_perms(node, user, W) and check_perms(old_parent, user, W) and check_perms(new_parent, user, W):
		files_table.update({"node_id": node_id}, {"$set": {"par_id": new_par_id}})

		old_entries = old_parent["entries"]
		old_entries.remove(node_id)
		files_table.update({"node_id": old_par_id}, {"$set": {"entries": old_entries}})

		new_entries = new_parent["entries"]
		new_entries.append(node_id)
		files_table.update({"node_id": new_par_id}, {"$set": {"entries": new_entries}})

	else:
		raise PermissionDeniedError

def cp_node(user, node_id, new_par_id):
	node = lookup_node(node_id)

	old_par_id = node["par_id"]
	old_parent = lookup_node(old_par_id)

	if old_parent == None:
		raise CorruptedDatabaseError

	new_parent = lookup_node(new_par_id)

	if node == None or new_parent == None:
		raise InvalidArgsError

	if check_perms(node, user, W) and check_perms(old_parent, user, R) and check_perms(new_parent, user, W):
		new_node_id = create_node(user, user, node["perms"][:2] + "0", new_par_id)
		os.system("cp " + node_id + " " + new_node_id)

		return new_node_id
	else:
		raise PermissionDeniedError

def init(mc):
	global mongo_client, files_table, login_db, users_table, groups_table
	
	mongo_client = mc

	files_table = mongo_client.tree.files

	login_db = mongo_client.login
	users_table = login_db.users
	groups_table = login_db.groups

	if lookup_node("0") is None:
		files_table.insert({"node_id": "0", "perms": "600", "owner": "root", "group": "root", "entries": [], "par_id": 0})

	printtree("0")
