import hashlib, os, time, uuid, json
from pymongo.errors import PyMongoError

from efs.common.encryption import generate_key
from efs.server.exceptions import ExistingGroupError
from efs.server.exceptions import ExistingUserError
from efs.server.exceptions import GroupMembershipError
from efs.server.exceptions import InvalidCredentialsError
from efs.server.exceptions import InvalidGroupError
from efs.server.exceptions import InvalidUserError
from efs.server import fs

class LoginService(object):

    def __init__(self, mongo_client):
        self.mongo_client = mongo_client
        try:
            self.login_db = mongo_client.login
            self.users = self.login_db.users
            self.sessions = self.login_db.sessions
            self.groups = self.login_db.groups
        except:
            raise PyMongoError

    def register(self, user, password, public_key):
        other_user = self.users.find({"user":user})
        if other_user.count() == 0:
            new_user = {}
            new_user["user"] = user
            new_user["salt"] = uuid.uuid4().hex
            new_user["password_hash"] = hashlib.sha512(password + new_user["salt"]).hexdigest()
            new_user["public_key"] = public_key
            self.users.insert(new_user)

            # New home directory
            return fs.create_node(user, user, "660", "0")
        else:
            raise ExistingUserError

    def login(self, user, password):
        user_record = self.users.find({"user":user}).limit(1)
        if user_record.count() == 1:
            password_hash = hashlib.sha512(password + user_record[0]["salt"]).hexdigest()
            if password_hash == user_record[0]["password_hash"]:
                session = self.generate_session(user)
                return session
            else:
                raise InvalidCredentialsError
        else:
            raise InvalidCredentialsError

    def logout(self, user):
        self.sessions.remove({"user": user})

    def change_password(self, user, new_password):
        user_record = self.users.find({"user":user})
        if user_record.count() == 1:
            new_password_hash = hashlib.sha512(new_password + user_record[0]["salt"]).hexdigest()
            self.users.update({"user":user}, {'$set':{"password_hash": new_password_hash}})
        else:
            raise InvalidUserError

    def make_group(self, group, owner, public_key, owner_private_key):
        if not group.isalnum():
            raise InvalidGroupError

        group_record = self.groups.find({"group":group})
        if group_record.count() == 0:
            new_group = {}
            new_group["group"] = group
            new_group["owner"] = owner
            new_group["public_key"] = public_key
            new_group["owner_private_key"] = owner_private_key
            new_group["members"] = {}
            self.groups.insert(new_group)
        else:
            raise ExistingGroupError

    def rm_group(self, group):
        try:
            self.groups.remove({"group":group})
        except:
            raise InvalidGroupError

    def get_groups(self, user):
        group_records = self.groups.find({})
        groups = []
        for i in range(group_records.count()):
            if user in group_records[i]["members"]:
                groups.append(group_records[i]["group"])
        return groups


    def add_user(self, user, group, member_private_key):
        group_record = self.groups.find({"group":group})
        if group_record.count() == 1:
            user_record = self.users.find({"user":user})
            if user_record.count() == 1:
                members = group_record[0]["members"]
                members[user] = member_private_key
                self.groups.update({"group":group}, {"$set":{"members":members}})
            else:
                raise InvalidUserError
        else:
            raise InvalidGroupError

    def rm_user(self, user, group):
        group_record = self.groups.find({"group":group})
        if group_record.count() == 1:
            user_record = self.users.find({"user":user})
            if user_record.count() == 1:
                members = group_record[0]["members"]
                if user in members:
                    del members[user]
                    self.groups.update({"group":group}, {"$set":{"members":members}})
                else:
                    raise GroupMembershipError
            else:
                raise InvalidUserError
        else:
            raise InvalidGroupError

    def generate_session(self, user):
        self.sessions.remove({"user":user})
        token_dict = {}
        token_dict["user"] = user
        token_dict["timestamp"] = int(time.time())
        token_dict["random"] = repr(os.urandom(32))
        new_session = {}
        new_session["user"] = user
        new_session["token"] = hashlib.sha512(json.dumps(token_dict)).hexdigest()
        #generate session key
        new_session["session_key"] = generate_key()
        new_session["timestamp"] = int(time.time())
        self.sessions.insert(new_session)
        return new_session

    def update_timestamp(self, user, timestamp):
        session_record = self.sessions.find({"user":user})
        if session_record.count() == 1 and session_record[0]["timestamp"] < timestamp:
            self.sessions.update({"user":user}, {"$set":{"timestamp":timestamp}})
